#include "wall_construction.h"

#include <argos3/plugins/simulator/entities/cylinder_entity.h>

#include <algorithm>
#include <cstring>
#include <cerrno>

/****************************************/
/****************************************/

static const Real OBJECT_RADIUS            = 0.1f;
static const Real OBJECT_DIAMETER          = OBJECT_RADIUS * 2.0f;
static const Real CONSTRUCTION_AREA_MIN_X  = 4.5f;
static const Real CONSTRUCTION_AREA_MAX_X  = 5.2f;
static const Real CONSTRUCTION_AREA_SIZE_X = CONSTRUCTION_AREA_MAX_X - CONSTRUCTION_AREA_MIN_X;
static const Real CONSTRUCTION_AREA_SIZE_Y = 4.0f - 0.2f - 0.2f; // arena size_y - west_wall size_y - east_wall size_y

/****************************************/
/****************************************/

CWallConstruction::CWallConstruction() {
}

/****************************************/
/****************************************/

CWallConstruction::~CWallConstruction() {
   /* Nothing to do */
}

/****************************************/
/****************************************/

void CWallConstruction::Init(TConfigurationNode& t_tree) {
   /* Get output file name from XML tree */
   GetNodeAttribute(t_tree, "output", m_strOutFile);
   /* Open the file for text writing */
   m_cOutFile.open(m_strOutFile.c_str(), std::ofstream::out | std::ofstream::trunc);
   if(m_cOutFile.fail()) {
      THROW_ARGOSEXCEPTION("Error opening file \"" << m_strOutFile << "\": " << ::strerror(errno));
   }
   /* Write a header line */
   m_cOutFile << "# Clock\tObjects\tThinness\tCompactness\tCoverage\t" << std::endl;
}

/****************************************/
/****************************************/

void CWallConstruction::Reset() {
   /* Close the output file */
   m_cOutFile.close();
   if(m_cOutFile.fail()) {
      THROW_ARGOSEXCEPTION("Error closing file \"" << m_strOutFile << "\": " << ::strerror(errno));
   }
   /* Open the file for text writing */
   m_cOutFile.open(m_strOutFile.c_str(), std::ofstream::out | std::ofstream::trunc);
   if(m_cOutFile.fail()) {
      THROW_ARGOSEXCEPTION("Error opening file \"" << m_strOutFile << "\": " << ::strerror(errno));
   }
   /* Write a header line */
   m_cOutFile << "# Clock\tObjects\tThinness\tCompactness\tCoverage\t" << std::endl;
}

/****************************************/
/****************************************/

void CWallConstruction::Destroy() {
   /* Close the output file */
   m_cOutFile.close();
   if(m_cOutFile.fail()) {
      THROW_ARGOSEXCEPTION("Error closing file \"" << m_strOutFile << "\": " << ::strerror(errno));
   }
}

/****************************************/
/****************************************/

void CWallConstruction::PreStep() {
   /* Nothing to do */
}

/****************************************/
/****************************************/

void CWallConstruction::PostStep() {
   FilterObjects();
   /* Output a line for this step */
   m_cOutFile << GetSpace().GetSimulationClock() << "\t"
              << m_vecConstructionObjectsInArea.size() << "\t"
              << CalculateThinness() << "\t"
              << CalculateCompactness() << "\t"
              << CalculateCoverage() << "\t"
              << std::endl;
}

/****************************************/
/****************************************/

CColor CWallConstruction::GetFloorColor(const CVector2& c_position_on_plane) {
   /* Check if the given point is within the construction area */
   if(c_position_on_plane.GetX() >= CONSTRUCTION_AREA_MIN_X &&
      c_position_on_plane.GetX() <= CONSTRUCTION_AREA_MAX_X) {
      /* Yes, it is - return black */
      return CColor::BLACK;
   }
   /* No, it isn't - return light gray */
   return CColor::GRAY90;
}

/****************************************/
/****************************************/

bool ObjectYCoordCompare(const CVector3& c_vec1,
                         const CVector3& c_vec2) {
   return c_vec1.GetY() < c_vec2.GetY();
}

void CWallConstruction::FilterObjects() {
   /* Clear list of positions of objects in construction area */
   m_vecConstructionObjectsInArea.clear();
   /* Clear values for the min and max Y coordinates */
   m_fMinObjectY =  100.0f;
   m_fMaxObjectY = -100.0f;
   /* Get the list of cylinders from the ARGoS space */
   CSpace::TMapPerType& tCylinderMap = GetSpace().GetEntitiesByType("cylinder");
   /* Go through the list and collect data */
   for(CSpace::TMapPerType::iterator it = tCylinderMap.begin();
       it != tCylinderMap.end();
       ++it) {
      /* Get a reference to the object body */
      CEmbodiedEntity& cBody = any_cast<CCylinderEntity*>(it->second)->GetEmbodiedEntity();
      /* Check if object is in target area */
      if(cBody.GetOriginAnchor().Position.GetX() > CONSTRUCTION_AREA_MIN_X &&
         cBody.GetOriginAnchor().Position.GetX() < CONSTRUCTION_AREA_MAX_X) {
         /* Yes, it is */
         /* Add it to the list */
         m_vecConstructionObjectsInArea.push_back(cBody.GetOriginAnchor().Position);
         /* Update min and max Y coordinates */
         m_fMinObjectY = Min(m_fMinObjectY, cBody.GetOriginAnchor().Position.GetY());
         m_fMaxObjectY = Max(m_fMaxObjectY, cBody.GetOriginAnchor().Position.GetY());
      }
   }
   /* Order vector elements by Y coordinate */
   std::sort(m_vecConstructionObjectsInArea.begin(),
             m_vecConstructionObjectsInArea.end(),
             ObjectYCoordCompare);
}

/****************************************/
/****************************************/

Real CWallConstruction::CalculateThinness() {
   /* There's no point proceeding if no objects are in the target area */
   if(m_vecConstructionObjectsInArea.empty()) {
      return 0.0f;
   }
   /* Calculate average of the X coordinate */
   Real fAverageX = 0.0f;
   for(size_t i = 0; i < m_vecConstructionObjectsInArea.size(); ++i) {
      fAverageX += m_vecConstructionObjectsInArea[i].GetX();
   }
   fAverageX /= static_cast<Real>(m_vecConstructionObjectsInArea.size());
   /* Calculate variance of the X coordinate */
   Real fVarianceX = 0.0f;
   for(size_t i = 0; i < m_vecConstructionObjectsInArea.size(); ++i) {
      fVarianceX += Square(fAverageX - m_vecConstructionObjectsInArea[i].GetX());
   }
   fVarianceX /= static_cast<Real>(m_vecConstructionObjectsInArea.size());
   /* Return the thinness */
   return 1.0f - Sqrt(fVarianceX) / CONSTRUCTION_AREA_SIZE_X;
}

/****************************************/
/****************************************/

Real CWallConstruction::CalculateCompactness() {
   /* There's no point proceeding if at least two objects are not in the target area */
   if(m_vecConstructionObjectsInArea.size() < 2) {
      return 0.0;
   }
   /* Calculate max of the distances */
   Real fMaxDistanceY = -100.0f;
   for(size_t i = 1; i < m_vecConstructionObjectsInArea.size(); ++i) {
      fMaxDistanceY = Max(fMaxDistanceY,
                          m_vecConstructionObjectsInArea[i].GetY() - m_vecConstructionObjectsInArea[i-1].GetY());
   }
   /* Calculate compactness */
   return fMaxDistanceY / OBJECT_DIAMETER;
}

/****************************************/
/****************************************/

Real CWallConstruction::CalculateCoverage() {
   /* There's no point proceeding if no objects are in the target area */
   if(m_vecConstructionObjectsInArea.empty()) {
      return 0.0;
   }
   /*
    * Calculate coverage
    */
   /* Start with coverage by the leftmost object */
   Real fCoverage = OBJECT_DIAMETER;
   Real fDistanceY;
   /* Go through further objects */
   for(size_t i = 1; i < m_vecConstructionObjectsInArea.size(); ++i) {
      /* Add coverage by this object */
      fCoverage += OBJECT_DIAMETER;
      /* Calculate distance between this object and previous objects */
      fDistanceY = m_vecConstructionObjectsInArea[i].GetY() - m_vecConstructionObjectsInArea[i-1].GetY();
      /* If the two objects overlap on Y axis, remove the overlap */
      if(fDistanceY < OBJECT_DIAMETER) {
         fCoverage -= OBJECT_DIAMETER - fDistanceY;
      }
   }
   return fCoverage / CONSTRUCTION_AREA_SIZE_Y;
}

/****************************************/
/****************************************/

/* Register this loop functions into the ARGoS plugin system */
REGISTER_LOOP_FUNCTIONS(CWallConstruction, "wall_construction");
