#include <argos3/core/simulator/loop_functions.h>
#include <argos3/plugins/robots/foot-bot/simulator/footbot_entity.h>
#include <fstream>

using namespace argos;

class CWallConstruction : public CLoopFunctions {

public:

   /**
    * Class constructor
    */
   CWallConstruction();

   /** 
    * Class destructor
    */
   virtual ~CWallConstruction();

   /**
    * Initializes the experiment.
    * It is executed once at the beginning of the experiment, i.e., when ARGoS is launched.
    * @param t_tree The parsed XML tree corresponding to the <loop_functions> section.
    */
   virtual void Init(TConfigurationNode& t_tree);

   /**
    * Resets the experiment to the state it was right after Init() was called.
    * It is executed every time you press the 'reset' button in the GUI.
    */
   virtual void Reset();

   /**
    * Undoes whatever Init() did.
    * It is executed once when ARGoS has finished the experiment.
    */
   virtual void Destroy();

   /**
    * Performs actions right before a simulation step is executed.
    */
   virtual void PreStep();

   /**
    * Performs actions right after a simulation step is executed.
    */
   virtual void PostStep();

   /**
    * Returns the color of the floor at the specified point on.
    * @param c_position_on_plane The position at which you want to get the color.
    * @see CColor
    */
   virtual CColor GetFloorColor(const CVector2& c_position_on_plane);

private:

   /**
    * Fills into m_vecConstructionObjectsInArea, and calculates m_fMin and m_fMax
    */
   void FilterObjects();

   /**
    * Calculates the thinness.
    * Thinness is defined as: 1 - A / B
    * where:
    *   A is the standard deviation of the X coordinates of the objects within
    *     the construction area
    *   B is the size along the X axis of the construction area (B = 0.5 meters)
    * When there are no objects in the construction area, thinness is
    * set to zero.
    */
   Real CalculateThinness();

   /**
    * Calculates the compactness.
    * The compactness is defined as: Gy / D
    * where:
    *   Gy is the maximum distance along the Y axis among the objects within
    *      the construction area
    *   D is the diameter of a construction object (D = 0.2 meters)
    * When there are less than 2 objects in the construction area, compactness is
    * set to zero.
    */
   Real CalculateCompactness();

   /**
    * Calculates the coverage.
    * The coverage is defined as: C / S
    * where:
    *   C is the projection of all the objects in the
    *     construction area on the Y axis.
    *   S is the size of the construction area along the Y axis
    * When there are no objects in the construction area, coverage is
    * set to zero.
    */
   Real CalculateCoverage();

private:

   /**
    * The path of the output file.
    */
   std::string m_strOutFile;

   /**
    * The stream associated to the output file.
    */
   std::ofstream m_cOutFile;

   /**
    * This vector contains a list of positions of objects in the construction area
    */
   std::vector<CVector3> m_vecConstructionObjectsInArea;

   /**
    * Minimum and maximum Y coordinate for the objects in the construction area
    */
   Real m_fMinObjectY, m_fMaxObjectY;

};
